# -*- coding: utf-8 -*-
"""
Contains a bunch of information about this package.
"""

__version__ = "0.1.5"

__author__ = "Benjamin F. Maier"
__copyright__ = "Copyright 2020-2021, " + __author__
__credits__ = [__author__]
__license__ = "MIT"
__maintainer__ = __author__
__email__ = "contact@benmaier.org"
__status__ = "Development"
